# SPDX-License-Identifier: MIT

import logging
import os
import subprocess

from time import sleep

log = logging.getLogger("calligrabot.signer")


def run_command(cmd, error_msg="", retries=6, backoff_factor=3, **kwargs):
    """
    Wrap around _run_command to throw an exception when the
    command fails to run and set up retries.

    :param cmd: Command to run.
    :param error_msg: Custom error message to show upon failure.
    :param retries: How many times to retry during a failure, 0 for no retries.
    :param backoff_factor: Number of seconds to use for exponential backoff.
    """
    sleep_time = 0
    for i in range(retries + 1):
        sleep(sleep_time)
        log.debug(
            "Executing command (Attempt #{}): {}".format(str(i + 1), cmd)
        )
        ret, stdout, stderr = _run_command(cmd, **kwargs)
        if stdout:
            log.debug("stdout: {}".format(stdout))

        if ret == 0:
            return ret, stdout, stderr

        if stderr:
            log.debug("stderr: {}".format(stderr))

        sleep_time = backoff_factor ** i
    else:
        if not error_msg:
            error_msg = "Failed to run command: {}. ".format(cmd)

        error_msg += "stderr: {}".format(stderr)
        raise Exception(error_msg)


def _run_command(cmd, **kwargs):
    """
    Run a command.

    :param cmd: Command to run.
    :param kwargs: Additional keyword arguments.
    :return: Tuple return code, stdout, stderr.
    """

    child = subprocess.Popen(
        cmd,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        **kwargs
    )
    stdout, stderr = child.communicate()
    ret = child.wait()
    return ret, stdout, stderr


def _build_koji_cmd(koji_profile):
    """
    Build command line to run koji with specified koji instance.

    :param koji_profile: Properties describing configured koji instance.
    :return: List containing command to invoke koji.
    """

    return [
        "/usr/bin/koji",
        "--server={}".format(koji_profile["server"]),
        "--weburl={}".format(koji_profile["weburl"]),
        "--topurl={}".format(koji_profile["topurl"]),
    ]


def download_build_rpm(koji_profile, rpm, path):
    """
    Download one rpm from koji.

    :param koji_profile: Properties describing configured koji instance.
    :param rpm: name-version-release.arch RPM name (without file extension).
    :param path: Directory path to which the RPM file should be downloaded.
    :returns: Full path to the downloaded RPM.
    """
    log.debug("Downloading {} from koji".format(rpm))

    # Robosignatory hands us a list of packages in this format: name-version-release.dist.arch
    # We need to make sure that we include the '.rpm' extension when we try to download from koji
    rpmbasename = rpm + '.rpm'
    koji_cmd = _build_koji_cmd(koji_profile)
    koji_cmd.extend(
        [
            "download-build",
            "--rpm",
            "--debuginfo",
            "--noprogress",
            rpmbasename,
        ]
    )
    _, stdout, _ = run_command(koji_cmd, cwd=path)

    rpmfile = os.path.join(path, rpmbasename)

    if not os.path.isfile(rpmfile):
        raise Exception("Downloaded RPM file {} is missing.".format(rpmfile))

    log.debug("Downloaded {} successfully".format(rpmfile))

    return rpmfile


def import_rpm_signatures(koji_profile, rpmfiles, dry_run=False):
    """
    Import signed rpms to koji.

    :param koji_profile: Properties describing configured koji instance.
    :param rpmfiles: List of RPM files to upload
    :param dry_run: True to make no changes
    :returns:
    """
    log.debug("Importing RPM signatures from {} to koji".format(rpmfiles))

    koji_cmd = _build_koji_cmd(koji_profile)
    koji_cmd.extend(["import-sig"])
    if dry_run:
        koji_cmd.extend(["--test"])
    koji_cmd.extend(rpmfiles)
    _, stdout, _ = run_command(koji_cmd)

    log.debug("Import of RPM signatures successful")


def generate_signing_cmdline(
    keyname,
    files_to_sign,
    principal=None,
    keytab=None,
    gpgsign=False,
    imasign=False,
    output_file=None,
    onbehalf="",
    nat=False,
    verbose=False,
    *args
):
    """
    Build the signing command to run.

    :param keyname: Name of the signing key.
    :param files_to_sign: List of the filenames of the RPM builds.
    :param principal: Kerberos principal for host authentication.
    :param keytab: Kerberos keytab for host authentication.
    :param gpgsign: Indicates whether it is a gpg/clear sign. Defaulted to
     False, which is for rpm sign.
    :param imasign: Indicates whether it is an IMA sign. Defaulted to
     False, which is for regular sign.
    :param output_file: Path to a file to send output to.
    :param onbehalf: Person who is triggered the signing.
    :param nat: True if application runs behind a NAT.
    :param verbose: True to run signing command with increased verbosity.
    :param args: List of extra args to append to signing command line.
    :return: List containing command line to invoke signing.
    """
    cmdline = ["rpm-sign"]

    if verbose:
        cmdline.extend(["--verbose"])

    cmdline.extend(["--key", keyname])
    if imasign:
        cmdline.extend(["--imasign"])
    cmdline.extend(files_to_sign)

    if not principal:
        raise ValueError("Required Kerberos principal is missing")
    if not keytab:
        raise ValueError("Required Kerberos keytab is missing")
    cmdline.extend(["--principal", principal, "--keytab", keytab])

    if gpgsign:
        cmdline.extend(["--gpgsign"])

    if output_file:
        cmdline.extend(["--output", output_file])

    if not onbehalf:
        raise ValueError("Required requester information is missing")
    cmdline.extend(["--onbehalfof", onbehalf])

    if nat:
        cmdline.extend(["--nat"])

    cmdline.extend(args)

    return cmdline
