# SPDX-License-Identifier: MIT

import click
import logging
import os
import rpm_head_signing
import tempfile

from fedora_messaging.config import conf

from calligrabot.utils import (
    download_build_rpm,
    generate_signing_cmdline,
    import_rpm_signatures,
    run_command,
)

log = logging.getLogger("calligrabot")


class UserObject:
    pass


@click.group()
@click.option(
    "-d",
    "--debug",
    is_flag=True,
    help="Enable debugging output",
)
@click.option(
    "-n",
    "--dry-run",
    is_flag=True,
    help="Dry run mode",
)
@click.option(
    "-c",
    "--config-file",
    default="/etc/fedora-messaging/robosignatory.toml",
    type=click.Path(exists=True, dir_okay=False),
    help="Path to the configuration file",
)
@click.option(
    "-u", "--user-name", required=True, type=click.STRING, help="User name"
)
@click.option(
    "-p", "--principal", type=click.STRING, help="Kerberos principal"
)
@click.option("-k", "--keytab", type=click.STRING, help="Kerberos keytab")
@click.pass_context
def cli(ctx, dry_run, debug, config_file, user_name, principal, keytab):
    if debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    ctx.obj = UserObject()

    if not os.path.isfile(config_file):
        raise click.exceptions.BadParameter(
            "{} is not a file".format(config_file)
        )

    conf.load_config(config_path=config_file)
    ctx.obj.cfg = conf["consumer_config"]

    ctx.obj.args = {}
    ctx.obj.args["debug"] = debug
    ctx.obj.args["dry_run"] = dry_run
    ctx.obj.args["user_name"] = user_name
    ctx.obj.args["principal"] = principal
    ctx.obj.args["keytab"] = keytab


@cli.command("sign-rpms")
@click.option(
    "-i",
    "--koji-instance",
    default="primary",
    help="Use the specified Koji instance",
)
@click.option(
    "-I", "--ima-sign-key", type=click.STRING, help="IMA signing key"
)
# Signing key name
@click.argument("keyname", nargs=1)
# List of RPM NEVRAs"
@click.argument("rpms", nargs=-1)
@click.pass_context
def sign_rpms(
    ctx,
    keyname,
    rpms,
    koji_instance,
    ima_sign_key,
):
    # print a bunch of debugging information
    # to make sure we have everything we need
    log.debug("sign_rpms() called")
    log.debug("keyname: {}".format(keyname))
    log.debug("rpms: {}".format(rpms))
    log.debug("koji_instance: {}".format(koji_instance))
    log.debug("ima_sign_key: {}".format(ima_sign_key))
    log.debug("debug: {}".format(ctx.obj.args["debug"]))
    log.debug("user_name: {}".format(ctx.obj.args["user_name"]))

    koji_profile = {
        "instance": koji_instance,
        "server": ctx.obj.cfg["koji_instances"][koji_instance]["url"],
        "weburl": ctx.obj.cfg["koji_instances"][koji_instance]["weburl"],
        "topurl": ctx.obj.cfg["koji_instances"][koji_instance]["topurl"],
    }
    log.debug("koji server: {}".format(koji_profile["server"]))
    log.debug("koji weburl: {}".format(koji_profile["weburl"]))
    log.debug("koji topurl: {}".format(koji_profile["topurl"]))

    # if principal or keytab haven't been provided on the command line,
    # use the values from the config file
    principal = ctx.obj.args["principal"] or ctx.obj.cfg["signing"].get(
        "principal"
    )
    keytab = ctx.obj.args["keytab"] or ctx.obj.cfg["signing"].get("keytab")
    log.debug("principal: {}".format(principal))
    log.debug("keytab: {}".format(keytab))

    tmpdir = tempfile.TemporaryDirectory(prefix="calligrabot-")
    log.info(
        "Downloading RPMs {} to temporary directory {}".format(
            rpms, tmpdir.name
        )
    )

    if ima_sign_key:
        imadigest = os.path.join(tmpdir.name, "ima_digests")

    rpmfiles = list()
    for rpm in rpms:
        rpmfile = download_build_rpm(koji_profile, rpm, tmpdir.name)
        rpmfiles.append(rpmfile)
        if ima_sign_key:
            rpm_head_signing.extract_header(
                rpmfile, rpmfile + ".hdr", imadigest
            )

    log.info(
        "Signing RPM files {} with key {}.".format(
            rpmfiles,
            keyname,
        )
    )
    sign_cmd = generate_signing_cmdline(
        keyname,
        rpmfiles,
        principal=principal,
        keytab=keytab,
        onbehalf=ctx.obj.args["user_name"],
        verbose=ctx.obj.args["debug"],
    )

    if ctx.obj.args["dry_run"]:
        log.info(
            "Dry run mode, not running signing command: {}".format(sign_cmd)
        )
    else:
        log.info("Running signing command: {}".format(sign_cmd))
        run_command(sign_cmd)

    if ima_sign_key:
        log.info(
            "IMA signing header digest {} with key {}.".format(
                imadigest,
                ima_sign_key,
            )
        )
        sign_cmd = generate_signing_cmdline(
            ima_sign_key,
            [imadigest],
            principal=principal,
            keytab=keytab,
            imasign=True,
            onbehalf=ctx.obj.args["user_name"],
            verbose=ctx.obj.args["debug"],
        )

        if ctx.obj.args["dry_run"]:
            log.info(
                "Dry run mode, not running IMA signing command: {}".format(
                    sign_cmd
                )
            )
        else:
            log.info("Running IMA signing command: {}".format(sign_cmd))
            run_command(sign_cmd)

        for rpmfile in rpmfiles:
            if ctx.obj.args["dry_run"]:
                log.info(
                    "Dry run mode, cannot insert IMA signatures into %s"
                    % rpmfile
                )
            else:
                log.info("Inserting IMA signatures into %s" % rpmfile)
                rpm_head_signing.insert_signature(
                    rpmfile,
                    None,
                    ima_presigned_path=imadigest + ".signed",
                )

    log.info("Importing RPM signatures to koji")

    import_rpm_signatures(
        koji_profile, rpmfiles, dry_run=ctx.obj.args["dry_run"]
    )
